;;;; complex-base.scm -- support for complex numbers in chicken
;;;; Copyright (c) 2004 by Thomas Christian Chust <chust@web.de>.
;;;; All rights reserved.
;;;;
;;;; Redistribution and use in source and binary forms, with or
;;;; without modification, are permitted provided that the following
;;;; conditions are met:
;;;;
;;;; * Redistributions of source code must retain the above copyright
;;;;   notice, this list of conditions and the following disclaimer.
;;;; * Redistributions in binary form must reproduce the above
;;;;   copyright notice, this list of conditions and the following
;;;;   disclaimer in the documentation and/or other materials provided
;;;;   with the distribution.
;;;; * Neither the name of the author nor the names of its
;;;;   contributors may be used to endorse or promote products derived
;;;;   from this software without specific prior written permission.
;;;;
;;;; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
;;;; CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
;;;; INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
;;;; MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
;;;; DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR
;;;; CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
;;;; SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
;;;; LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
;;;; USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
;;;; AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
;;;; LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
;;;; ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
;;;; POSSIBILITY OF SUCH DAMAGE.

(cond-expand
 ((or chicken-compile-shared shared)
  )
 (else
  (declare (unit complex))))

(include "complex.scm")

(declare (export
	 real-sqrt real-expt
	 real-abs

	 make-rectangular
	 make-polar
	 real-part
	 imag-part
	 magnitude
	 abs
	 angle

	 complex? number?
	 true-complex? imaginary?
	 real? rational? integer?
	 exact? inexact?
	 zero? positive? odd? even?
	 = < > <= >= max min
	 + - * /
	 quotient remainder modulo
	 gcd lcm
	 floor ceiling truncate round
	 exact->inexact inexact->exact
	 number->string string->number
	 exp log
	 sin cos tan
	 asin acos atan
	 sqrt expt
	 conj
	 i -i j -j))


(require-extension srfi-1 srfi-9 srfi-10 extras regex-case)

;; signal a type error with the given message
(define (type-error msg)
  (abort (make-composite-condition
	  (make-property-condition 'exn 'message msg)
	  (make-property-condition 'type))))

;; some procedures need to be saved with new names
(define real-real? real?)
(define real-rational? rational?)
(define real-integer? integer?)
(define real-exact? exact?)
(define real-inexact? inexact?)
(define real-zero? zero?)
(define real-positive? positive?)
(define real-odd? odd?)
(define real-even? even?)
(define real= =)
(define real< <)
(define real> >)
(define real<= <=)
(define real>= >=)
(define real+ +)
(define real- -)
(define real* *)
(define real/ /)
(define real-quotient quotient)
(define real-remainder remainder)
(define real-modulo modulo)
(define real-gcd gcd)
(define real-lcm lcm)
(define real-floor floor)
(define real-ceiling ceiling)
(define real-truncate truncate)
(define real-round round)
(define real-exact->inexact exact->inexact)
(define real-inexact->exact inexact->exact)
(define real-number->string number->string)
(define real-string->number string->number)
(define real-exp exp)
(define real-log log)
(define real-sin sin)
(define real-cos cos)
(define real-tan tan)
(define real-asin asin)
(define real-acos acos)
(define real-atan atan)
(define real-sqrt sqrt)
(define real-expt expt)
(define real-abs abs)
(define real-max max)
(define real-min min)

;; define the data type
(define-record-type complex
  (make-complex real imag)
  complex-record?
  (real complex-real)
  (imag complex-imag))

;; and some publicly visible constructors
(define (make-rectangular real imag)
  (unless (and (real-real? real) (real-real? imag))
	  (type-error "real and imaginary part must be numbers"))
  (make-complex real imag))

(define-reader-ctor 'complex make-rectangular)

(define (make-polar r phi)
  (unless (and (real-real? r) (real-real? phi))
	  (type-error "r and phi must be numbers"))
  (make-complex (real* r (real-cos phi)) (real* r (real-sin phi))))

(define (real-part n)
  (if (real-real? n)
      n
      (complex-real n)))

(define (imag-part n)
  (if (real-real? n)
      0
      (complex-imag n)))

;; redefine numerical predicates
(define (complex? n)
  (or (real-real? n) (complex-record? n)))

(define (true-complex? n)
  (and (complex-record? n) (not (zero? (complex-imag n)))))

(define number? complex?)

(define (imaginary? n)
  (and (true-complex? n) (real-zero? (complex-real n))))

(define (real? n)
  (or (real-real? n)
      (and (complex-record? n)
	   (zero? (complex-imag n))
	   (not (zero? (complex-real n))))))

;; the first really useful functions
(define (magnitude n)
  (if (real? n)
      (real-abs (real-part n))
      (let ((x (real-part n)) (y (imag-part n)))
	(real-sqrt (real+ (real* x x) (real* y y))))))

(define abs magnitude)

(define (angle n)
  (real-atan (imag-part n) (real-part n)))

;; and more predicates
(define (rational? n)
  (or (and (real-real? n) (real-rational? n))
      (and (complex-record? n)
	   (real-rational? (complex-real n)) (real-zero? (complex-imag n)))))

(define (integer? n)
  (or (and (real-real? n) (real-integer? n))
      (and (complex-record? n)
	   (real-integer? (complex-real n)) (real-zero? (complex-imag n)))))

(define (exact? n)
  (or (and (real-real? n) (real-exact? n))
      (and (complex-record? n)
	   (real-exact? (complex-real n)) (real-exact? (complex-imag n)))))

(define (inexact? n)
  (or (and (real-real? n) (real-inexact? n))
      (and (complex-record? n)
	   (or (real-inexact? (complex-real n))
	       (real-inexact? (complex-imag n))))))

(define (zero? n)
  (or (and (real-real? n) (real-zero? n))
      (and (complex-record? n)
	   (real-zero? (complex-real n)) (real-zero? (complex-imag n)))))

(define (positive? n)
  (or (and (real-real? n) (real-positive? n))
      (and (complex-record? n)
	   (real-positive? (complex-real n)) (real-zero? (complex-imag n)))))

(define (odd? n)
  (or (and (real-real? n) (real-odd? n))
      (and (complex-record? n)
	   (real-odd? (complex-real n)) (real-zero? (complex-imag n)))))

(define (even? n)
  (or (and (real-real? n) (real-even? n))
      (and (complex-record? n)
	   (real-even? (complex-real n)) (real-zero? (complex-imag n)))))

;; a helper converting numbers to complex representation
;; whether they are true complex or not
(define (expand-to-complex ns)
  (map (lambda (n)
	 (if (real-real? n) (make-rectangular n 0) n))
       ns))

;; a helper converting a number to real if it's not true complex
(define (collapse-to-real n)
  (if (real? n) (real-part n) n))

;; define numeric comparison operators
(define (= . ns)
  (let ((cns (expand-to-complex ns)))
    (and (apply real= (map complex-real cns))
	 (apply real= (map complex-imag cns)))))

;; the following comparisons can only be applied to reals
(define (< . ns)
  (let ((cns (expand-to-complex ns)))
    (if (apply real= (cons 0 (map complex-imag cns)))
	(apply real< (map complex-real cns))
	(type-error "complex numbers have no relation of order"))))

(define (> . ns)
  (let ((cns (expand-to-complex ns)))
    (if (apply real= (cons 0 (map complex-imag cns)))
	(apply real> (map complex-real cns))
	(type-error "complex numbers have no relation of order"))))

(define (<= . ns)
  (let ((cns (expand-to-complex ns)))
    (if (apply real= (cons 0 (map complex-imag cns)))
	(apply real<= (map complex-real cns))
	(type-error "complex numbers have no relation of order"))))

(define (>= . ns)
  (let ((cns (expand-to-complex ns)))
    (if (apply real= (cons 0 (map complex-imag cns)))
	(apply real>= (map complex-real cns))
	(type-error "complex numbers have no relation of order"))))

(define (max . ns)
  (let ((cns (expand-to-complex ns)))
    (if (apply real= (cons 0 (map complex-imag cns)))
	(apply real-max (map complex-real cns))
	(type-error "complex numbers have no relation of order"))))

(define (min . ns)
  (let ((cns (expand-to-complex ns)))
    (if (apply real= (cons 0 (map complex-imag cns)))
	(apply real-min (map complex-real cns))
	(type-error "complex numbers have no relation of order"))))

;; numerical operations
(define (+ . ns)
  (collapse-to-real
   (reduce (lambda (p q)
	     (let ((a (complex-real p)) (b (complex-imag p))
		   (c (complex-real q)) (d (complex-imag q)))
	       (make-complex (real+ a c) (real+ b d))))
	   (make-rectangular 0 0)
	   (expand-to-complex ns))))

(define (* . ns)
  (collapse-to-real
   (reduce (lambda (p q)
	     (let ((a (complex-real p)) (b (complex-imag p))
		   (c (complex-real q)) (d (complex-imag q)))
	       (make-complex (real- (real* a c) (real* b d))
			     (real+ (real* a d) (real* b c)))))
	   (make-rectangular 1 0)
	   (expand-to-complex ns))))

(define (- . ns)
  (let ((cns (expand-to-complex ns)))
    (collapse-to-real
     (if (fx= (length cns) 1)
	 (make-complex (real- (complex-real (car cns)))
		       (real- (complex-imag (car cns))))
	 (fold (lambda (p q)
		 (let ((a (complex-real p)) (b (complex-imag p))
		       (c (complex-real q)) (d (complex-imag q)))
		   (make-complex (real- c a) (real- d b))))
	       (car cns)
	       (cdr cns))))))

(define (/ . ns)
  (let ((cns (expand-to-complex ns)))
    (collapse-to-real
     (if (fx= (length cns) 1)
	 (let* ((a (complex-real (car cns))) (b (complex-imag (car cns)))
		(r (real+ (real* a a) (real* b b))))
	   (make-complex (real/ a r) (real- (real/ b r))))
	 (fold (lambda (p q)
		 (let ((c (complex-real p)) (d (complex-imag p))
		       (a (complex-real q)) (b (complex-imag q)))
		   (let ((r (real+ (real* c c) (real* d d))))
		     (make-complex
		      (real/ (real+ (real* a c) (real* b d)) r)
		      (real/ (real- (real* b c) (real* a d)) r)))))
	       (car cns)
	       (cdr cns))))))

;; and again some arithmetics that can only be performed on reals
(define (quotient a b)
  (let ((cns (expand-to-complex (list a b))))
    (if (apply real= (cons 0 (map complex-imag cns)))
	(apply real-quotient (map complex-real cns))
	(type-error "complex numbers have no relation of order"))))

(define (remainder a b)
  (let ((cns (expand-to-complex (list a b))))
    (if (apply real= (cons 0 (map complex-imag cns)))
	(apply real-remainder (map complex-real cns))
	(type-error "complex numbers have no relation of order"))))

(define (modulo a b)
  (let ((cns (expand-to-complex (list a b))))
    (if (apply real= (cons 0 (map complex-imag cns)))
	(apply real-modulo (map complex-real cns))
	(type-error "complex numbers have no relation of order"))))

(define (gcd . ns)
  (let ((cns (expand-to-complex ns)))
    (if (apply real= (cons 0 (map complex-imag cns)))
	(apply real-gcd (map complex-real cns))
	(type-error "complex numbers have no relation of order"))))

(define (lcm . ns)
  (let ((cns (expand-to-complex ns)))
    (if (apply real= (cons 0 (map complex-imag cns)))
	(apply real-lcm (map complex-real cns))
	(type-error "complex numbers have no relation of order"))))

;; rounding procedures are applied independently to the real and
;; imaginary parts for complex numbers
(define (floor n)
  (collapse-to-real (make-complex (real-floor (real-part n))
				  (real-floor (imag-part n)))))

(define (ceiling n)
  (collapse-to-real (make-complex (real-ceiling (real-part n))
				  (real-ceiling (imag-part n)))))

(define (truncate n)
  (collapse-to-real (make-complex (real-truncate (real-part n))
				  (real-truncate (imag-part n)))))

(define (round n)
  (collapse-to-real (make-complex (real-round (real-part n))
				  (real-round (imag-part n)))))

;; conversions
(define (exact->inexact n)
  (collapse-to-real (make-complex (real-exact->inexact (real-part n))
				  (real-exact->inexact (imag-part n)))))

(define (inexact->exact n)
  (collapse-to-real (make-complex (real-inexact->exact (real-part n))
				  (real-inexact->exact (imag-part n)))))

(define (number->string n #!optional (radix 10))
  (if (real? n)
      (real-number->string n radix)
      (let ((real (real-part n)) (imag (imag-part n)))
	(sprintf "#,(complex ~a ~a)"
		 (real-number->string real radix)
		 (real-number->string (real-abs imag) radix)))))

;; now that we can print it, we should define a record printer
(define-record-printer (complex number port)
  (display (number->string number) port))

(define (string->number s #!optional (radix 10))
  (handle-exceptions
   x #f
   (regex-case s
     ("\\s*(\\+\\s*i|i)\\s*" _
      i)
     ("\\s*\\-\\s*i\\s*" _
      -i)
     ("\\s*e\\^\\s*([^\\s]+)\\s*i\\s*" (_ phi)
      (collapse-to-real
       (make-polar
	1 (real-string->number phi radix))))
     ("\\s*([^\\s]+)\\s*\\*\\s*e\\^\\s*([^\\s]+)\\s*i\\s*" (_ r phi)
      (collapse-to-real
       (make-polar
	(real-string->number r radix)
	(real-string->number phi radix))))
     ("\\s*([^\\s]+)\\s*\\+\\s*i\\s*" (_ real)
      (make-rectangular
       (real-string->number real radix)
       1))
     ("\\s*([^\\s]+)\\s*\\-\\s*i" (_ real)
      (make-rectangular
       (real-string->number real radix)
       -1))
     ("\\s*([^\\s]+)\\s*\\+\\s*([^\\s]+)\\s*i\\s*" (_ real imag)
      (collapse-to-real
       (make-rectangular
	(real-string->number real radix)
	(real-string->number imag radix))))
     ("\\s*([^\\s]+)\\s*\\-\\s*([^\\s]+)\\s*i\\s*" (_ real imag)
      (collapse-to-real
       (make-rectangular
	(real-string->number real radix)
	(real- (real-string->number imag radix)))))
     ("\\s*([^\\s]+)\\s*i\\s*" (_ imag)
      (collapse-to-real
       (make-rectangular
	0 (real-string->number imag radix))))
     ("([^\\s]+)\\s*\\s*([^\\s]+)" (_ real imag)
      (collapse-to-real
       (make-rectangular
	(real-string->number real radix)
	(real-string->number imag radix))))
     ("#,\\(complex\\s+([^\\s]+)\\s+([^\\s]+)\\)" (_ real imag)
      (collapse-to-real
       (make-rectangular
	(real-string->number real radix)
	(real-string->number imag radix))))
     (else (real-string->number s radix)))))

;; more complex mathematical functions
(define (exp n)
  (if (real? n)
      (real-exp (real-part n))
      (* (real-exp (real-part n))
	 (let ((p (imag-part n)))
	   (make-complex (real-cos p) (real-sin p))))))

(define (log n)
  (if (and (real? n) (>= n 0))
      (real-log (real-part n))
      (make-complex (real-log (abs n)) (angle n))))

(define (sin n)
  (if (real? n)
      (real-sin (real-part n))
      (let ((in (* i n)))
	(/ (- (exp in) (exp (- in))) (make-rectangular 0 2)))))

(define (cos n)
  (if (real? n)
      (real-cos (real-part n))
      (let ((in (* i n)))
	(/ (+ (exp in) (exp (- in))) (make-rectangular 2 0)))))

(define (tan n)
  (if (real? n)
      (real-tan (real-part n))
      (/ (sin n) (cos n))))

(define (asin n)
  (collapse-to-real
   (* -i (log (+ (* i n) (sqrt (- 1 (* n n))))))))

(define (acos n)
  (- (real-asin 1) (asin n)))

(define (atan n #!optional b)
  (cond
   ((and (real? n) (real? b))
    (real-atan (real-part n) (real-part b)))
   ((not b)
    (let ((in (* i n)))
      (/ (- (log (+ 1 in)) (log (- 1 in))) (make-rectangular 0 2))))
   (else
    (type-error "atan with two arguments needs real parameters"))))

(define (sqrt n)
  (* (real-sqrt (abs n))
     (let ((p (real/ (angle n) 2)))
       (make-complex (real-cos p) (real-sin p)))))

(define (expt a b)
  (if (and (real? a) (real? b) (>= a 0))
      (real-expt (real-part a) (real-part b))
      (exp (* b (log a)))))

;; what they forgot: the complex conjugate
(define (conj n)
  (collapse-to-real (make-complex (real-part n) (real- (imag-part n)))))

;; and those constants may be handy
(define i (make-rectangular 0 1))
(define -i (make-rectangular 0 -1))
(define j (make-rectangular 0 1))
(define -j (make-rectangular 0 -1))
